(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('vue')) :
  typeof define === 'function' && define.amd ? define(['exports', 'vue'], factory) :
  (global = global || self, factory(global.Vue = {}, global.Vue));
}(this, function (exports, Vue) { 'use strict';

  Vue = Vue && Vue.hasOwnProperty('default') ? Vue['default'] : Vue;

  var observable = Vue.observable;
  var nextTick = Vue.nextTick;

  var ValueWrapperSymbol = Symbol('SymbolValueWrapper');
  function isValueWrapper(source) {
    return source ? source[ValueWrapperSymbol] === 1 : false
  }

  var currentInstance = null;

  function setCurrentInstance (instance) {
    currentInstance = instance;
  }

  function ensureCurrentInstance() {
    if (!currentInstance) {
      throw new Error("invalid call.")
    }
  }

  Vue.mixin({
    beforeCreate: function beforeCreate() {
      var ref = this.$options;
      var setup = ref.setup;
      if (setup) {
        this.$options.setup = setup.bind(this);
      }
    },
    created: function created() {
      setCurrentInstance(this);
      var ref = this.$options;
      var setup = ref.setup;
      if (setup) {
        var result = setup(this.$props);
        if (typeof result === 'function') {
          this.$options.render = result;
        } else {
          Object.assign(this._self, result);
        }
      }
      setCurrentInstance(null);
    }
  });

  function value(initialVal) {
    var obj;

    return observable(( obj = { value: initialVal }, obj[ValueWrapperSymbol] = 1, obj ))
  }

  function computed(getter, setter) {
    var obj$1;

    var obj = Object.defineProperty(( obj$1 = {}, obj$1[ValueWrapperSymbol] = 1, obj$1 ), 'value', {
      get: getter,
      set: setter
    });
    return observable(obj)
  }

  // 1. function that returns some value
  // 2. a wrapper object
  // 3. Array of [1] or [2]
  function watch(source, callback, option) {
    if (!callback) { return }

    // TODO flush
    var ref = option || {};
    var deep = ref.deep; if ( deep === void 0 ) deep = false;
    var lazy = ref.lazy; if ( lazy === void 0 ) lazy = false;

    var cleanup = null;
    var immediate = !lazy;

    var isArray = Array.isArray(source);
    var setCleanUp = function (cb) { return (cleanup = cb); };
    var runCleanUp = function () { return typeof cleanup === 'function' && cleanup(); };

    var onChange = function (val, oldVal) {
      if (immediate && isArray) {
        oldVal = oldVal || new Array(source.length);
        immediate = false;
      }
      runCleanUp();
      // TODO call according to flush
      callback(val, oldVal, setCleanUp);
    };

    var isInSetup = !!currentInstance;
    var vm = isInSetup ? currentInstance : new Vue();

    var watchFn = null;
    if (Array.isArray(source)) {
      var sources = source.map(transform);
      var properties = sources.reduce(function (acc, getter, i) {
        acc[("_" + i)] = { get: getter };
        return acc
      }, {});
      var obj = Object.defineProperties({}, properties);
      var state = observable(obj);
      watchFn = function () { return sources.map(function (_, i) { return state['_' + i]; }); };
    } else {
      watchFn = transform(source);
    }

    var stopped = false;
    var stop = vm.$watch(watchFn, onChange, { deep: deep, immediate: !lazy });
    var unwatch = function () {
      if (!stopped) {
        runCleanUp();
        stop();
        stopped = true;

        if (!isInSetup) {
          vm.$destroy();
        }
      }
    };

    vm.$on('hook:destroyed', unwatch);

    return unwatch
  }

  function transform(source) {
    if (source) {
      if (isValueWrapper(source)) {
        return function () { return source.value; }
      }

      if (typeof source === 'function') {
        return source
      }
    }

    /* istanbul ignore next */
    throw new Error('Invalid source.')
  }

  var setHook = function (lifecycle, fn) {
    ensureCurrentInstance();
    currentInstance.$on('hook:' + lifecycle, fn);
  };

  // We cannot access props until created...
  // export function onBeforeCreate(fn) {
  //   setHook('beforeCreate', fn)
  // }

  function onCreated(fn) {
    // Since we inject `$options.setup` in `created` hook
    fn();
    // setHook('created', fn)
  }

  function onBeforeMount(fn) {
    setHook('beforeMount', fn);
  }

  function onMounted(fn) {
    setHook('mounted', fn);
  }

  function onBeforeUpdate(fn) {
    setHook('beforeUpdate', fn);
  }

  function onUpdated(fn) {
    setHook('updated', fn);
  }

  function onBeforeUnmount(fn) {
    setHook('beforeDestroy', fn);
  }

  function onUnmounted(fn) {
    setHook('destroyed', fn);
  }

  function onActivated(fn) {
    setHook('activated', fn);
  }

  function onDeactivated(fn) {
    setHook('deactivated', fn);
  }

  function onErrorCaptured(fn) {
    ensureCurrentInstance();
    var ref = currentInstance.$options;
    var errorCaptured = ref.errorCaptured;

    /* istanbul ignore next */
    currentInstance.$options.errorCaptured = errorCaptured
      ? [fn].concat(errorCaptured)
      : [fn];
  }

  function createComponent(vue) {
    return vue
  }

  exports.computed = computed;
  exports.createComponent = createComponent;
  exports.default = Vue;
  exports.nextTick = nextTick;
  exports.onActivated = onActivated;
  exports.onBeforeMount = onBeforeMount;
  exports.onBeforeUnmount = onBeforeUnmount;
  exports.onBeforeUpdate = onBeforeUpdate;
  exports.onCreated = onCreated;
  exports.onDeactivated = onDeactivated;
  exports.onErrorCaptured = onErrorCaptured;
  exports.onMounted = onMounted;
  exports.onUnmounted = onUnmounted;
  exports.onUpdated = onUpdated;
  exports.state = observable;
  exports.value = value;
  exports.watch = watch;

  Object.defineProperty(exports, '__esModule', { value: true });

}));
