import Vue from 'vue';

var observable = Vue.observable;
var nextTick = Vue.nextTick;

var ValueWrapperSymbol = Symbol('SymbolValueWrapper');
function isValueWrapper(source) {
  return source ? source[ValueWrapperSymbol] === 1 : false
}

var currentInstance = null;

function setCurrentInstance (instance) {
  currentInstance = instance;
}

function ensureCurrentInstance() {
  if (!currentInstance) {
    throw new Error("invalid call.")
  }
}

Vue.mixin({
  beforeCreate: function beforeCreate() {
    var ref = this.$options;
    var setup = ref.setup;
    if (setup) {
      this.$options.setup = setup.bind(this);
    }
  },
  created: function created() {
    setCurrentInstance(this);
    var ref = this.$options;
    var setup = ref.setup;
    if (setup) {
      var result = setup(this.$props);
      if (typeof result === 'function') {
        // TODO implementation: new render function
        // SEE https://github.com/vuejs/rfcs/blob/render-fn-api-change/active-rfcs/0000-render-function-api-change.md
        this.$options.render = result;
      } else {
        Object.assign(this._self, result);
      }
    }
    setCurrentInstance(null);
  }
});

function keys(obj) {
  if (typeof Reflect !== 'undefined') {
    return Reflect.ownKeys(obj)
  }
  return Object.getOwnPropertyNames(obj).concat( Object.getOwnPropertySymbols(obj)
  )
}

function state(o) {
  var obj = Object.assign({}, o);
  // According to the RFC ——
  // ```
  // Value wrappers are also automatically unwrapped when accessed as a nested
  // property inside a reactive object
  // ```
  keys(obj).forEach(function (key) {
    var val = obj[key];
    if (isValueWrapper(val)) {
      Object.defineProperty(obj, key, {
        get: function get() {
          return val.value
        },
        set: function set(v) {
          val.value = v;
        }
      });
    }
  });
  return observable(obj)
}

function value(initialVal) {
  var obj;

  return observable(( obj = { value: initialVal }, obj[ValueWrapperSymbol] = 1, obj ))
}

function computed(getter, setter) {
  var obj$1;

  var obj = Object.defineProperty(( obj$1 = {}, obj$1[ValueWrapperSymbol] = 1, obj$1 ), 'value', {
    get: getter,
    set: setter
  });
  return observable(obj)
}

// 1. function that returns some value
// 2. a wrapper object
// 3. Array of [1] or [2]
function watch(source, callback, option) {
  if (!callback) { return }

  // TODO flush
  var ref = option || {};
  var deep = ref.deep; if ( deep === void 0 ) deep = false;
  var lazy = ref.lazy; if ( lazy === void 0 ) lazy = false;

  var cleanup = null;
  var immediate = !lazy;

  var isArray = Array.isArray(source);
  var setCleanUp = function (cb) { return (cleanup = cb); };
  var runCleanUp = function () { return typeof cleanup === 'function' && cleanup(); };

  var onChange = function (val, oldVal) {
    if (immediate && isArray) {
      oldVal = oldVal || new Array(source.length);
      immediate = false;
    }
    runCleanUp();
    // TODO call according to flush
    callback(val, oldVal, setCleanUp);
  };

  var isInSetup = !!currentInstance;
  var vm = isInSetup ? currentInstance : new Vue();

  var watchFn = null;
  if (Array.isArray(source)) {
    var sources = source.map(transform);
    var properties = sources.reduce(function (acc, getter, i) {
      acc[("_" + i)] = { get: getter };
      return acc
    }, {});
    var obj = Object.defineProperties({}, properties);
    var state = observable(obj);
    watchFn = function () { return sources.map(function (_, i) { return state['_' + i]; }); };
  } else {
    watchFn = transform(source);
  }

  var stopped = false;
  var stop = vm.$watch(watchFn, onChange, { deep: deep, immediate: !lazy });
  var unwatch = function () {
    if (!stopped) {
      runCleanUp();
      stop();
      stopped = true;

      if (!isInSetup) {
        vm.$destroy();
      }
    }
  };

  vm.$on('hook:destroyed', unwatch);

  return unwatch
}

function transform(source) {
  if (source) {
    if (isValueWrapper(source)) {
      return function () { return source.value; }
    }

    if (typeof source === 'function') {
      return source
    }
  }

  /* istanbul ignore next */
  throw new Error('Invalid source.')
}

var setHook = function (lifecycle, fn) {
  ensureCurrentInstance();
  currentInstance.$on('hook:' + lifecycle, fn);
};

// We cannot access props until created...
// export function onBeforeCreate(fn) {
//   setHook('beforeCreate', fn)
// }

function onCreated(fn) {
  // Since we inject `$options.setup` in `created` hook
  fn();
  // setHook('created', fn)
}

function onBeforeMount(fn) {
  setHook('beforeMount', fn);
}

function onMounted(fn) {
  setHook('mounted', fn);
}

function onBeforeUpdate(fn) {
  setHook('beforeUpdate', fn);
}

function onUpdated(fn) {
  setHook('updated', fn);
}

function onBeforeUnmount(fn) {
  setHook('beforeDestroy', fn);
}

function onUnmounted(fn) {
  setHook('destroyed', fn);
}

function onActivated(fn) {
  setHook('activated', fn);
}

function onDeactivated(fn) {
  setHook('deactivated', fn);
}

function onErrorCaptured(fn) {
  ensureCurrentInstance();
  var ref = currentInstance.$options;
  var errorCaptured = ref.errorCaptured;

  /* istanbul ignore next */
  currentInstance.$options.errorCaptured = errorCaptured
    ? [fn].concat(errorCaptured)
    : [fn];
}

function createComponent(vue) {
  return vue
}

export default Vue;
export { computed, createComponent, nextTick, onActivated, onBeforeMount, onBeforeUnmount, onBeforeUpdate, onCreated, onDeactivated, onErrorCaptured, onMounted, onUnmounted, onUpdated, state, value, watch };
